﻿open System.Diagnostics
open System.Threading

(* Raima.Rdm is the system module that contains the ADO.NET data provider
   for RDM *)
open Raima.Rdm

(* EXAMPLE - SQL14 for F# using ADO.NET *)
(* This example shows multi-user access to a database using separate processes
 * and a remote tfserver.  The connections will use TCP/IP to communicate
 * with the tfserver to do their work.  After starting the tfserver, two
 * writer processes are created.  The first begins inserting values into the
 * table.  The second waits until the table contains more than 100 rows and
 * then begins deleting the oldest.  A reader process is also created that
 * prints out the current oldest and newest entries every second.  Each
 * process uses regular locking.
 * 
 * The following methods are illustrated:
 *     Raima.Rdm.RdmConnection:
 *         ChangeDatabase(string) Method
 *         Close() Method
 *         CreateCommand() Method
 *         Open() Method
 *         RdmConnection(String) Constructor
 *     Raima.Rdm.RdmCommand:
 *         CommandText Property
 *         ExecuteNonQuery() Method
 *         ExecuteReader() Method
 *         Parameters Property
 *     Raima.Rdm.RdmDataReader:
 *         Item[int] Property
 *         Read() Method
 *     Raima.Rdm.RdmParameterCollection
 *         Add(string, RdmType)
 *         Item[string] Property
 * 
 * They are fully document in http://docs.raima.com/rdm/.
 * 
 * For simplicity this example does not catch any thrown exceptions
 * but good programming practices would dictate that this occur in
 * a full application.
 *)

let DropDatabase (conn : RdmConnection) name =
    let cmd = conn.CreateCommand()

    (* Attempt to drop this database.  If the database does not already exist
     * this will fail and we simply ignore the failure for this sample. *)
    try
        cmd.CommandText <- "DROP DATABASE " + name
        cmd.ExecuteNonQuery() |> ignore
    with
        | exn -> ()

let CreateThisDatabase (conn : RdmConnection) name cmds =
    let cmd = conn.CreateCommand()

    cmd.CommandText <- "CREATE DATABASE " + name
    cmd.ExecuteNonQuery() |> ignore

    cmds |>
        Array.iter(fun c ->
            cmd.CommandText <- c
            cmd.ExecuteNonQuery() |> ignore
        )

    cmd.CommandText <- "COMMIT"
    cmd.ExecuteNonQuery() |> ignore

    cmd.CommandText <- "CLOSE " + name
    cmd.ExecuteNonQuery() |> ignore

let CreateDatabases conn =
    (* Drop any existing copies of the databases. *)
    DropDatabase conn "SQL14"

    (* Attempt to create the databases *)
    CreateThisDatabase conn "SQL14" [| "CREATE TABLE Table1 (id INTEGER PRIMARY KEY)" |]

let Inserter _ =
    printfn "Inserter process"

    (* Create a connection for this process *)
    let conn = new RdmConnection("host=local;database=SQL14;openmode=s;tfstype=tfsr")
    conn.Open()

    (* Create a command to do the insert *)
    let cmd = conn.CreateCommand()
    cmd.CommandText <- "INSERT INTO Table1 VALUES :id"
    cmd.Parameters.Add("id", RdmType.Int32) |> ignore

    (* Loop until a shutdown is requested *)
    let seq = Seq.initInfinite (fun ii -> ii)
    
    seq |>
        Seq.iter(fun ii ->
            (* Insert the next value *)
            cmd.Parameters.["id"].Value <- ii
            cmd.ExecuteNonQuery() |> ignore)
    
let Deleter _ =
    printfn "Deleter process"

    (* Create a connection for this process *)
    let conn = new RdmConnection("host=local;database=SQL14;openmode=s;tfstype=tfsr")
    conn.Open()

    (* Create a command to check the number of rows, one to find the
     * 'smallest' row, and one to delete *)
    let chkCmd = conn.CreateCommand()
    let minCmd = conn.CreateCommand()
    let delCmd = conn.CreateCommand()

    chkCmd.CommandText <- "SELECT COUNT(id) FROM Table1"
    minCmd.CommandText <- "SELECT MIN(id) FROM Table1"
    delCmd.CommandText <- "DELETE FROM Table1 WHERE id = :minid"
    let parm = delCmd.Parameters.Add("minid", RdmType.Int32)

    (* Loop until a shutdown is requested *)
    while true do
        (* Query the number of rows in the table *)
        let countRdr = chkCmd.ExecuteReader()
        if countRdr.Read() then
            let count = System.Convert.ToInt64(countRdr.[0])

            (* There should be only one row, so this closes the reader *)
            countRdr.Read() |> ignore

            if (count > 100L) then
                (* If there are at least 100 rows, get the smallest and delete it *)
                let minRdr = minCmd.ExecuteReader()
                if minRdr.Read() then
                    parm.Value <- System.Convert.ToInt32(minRdr.[0])
                    delCmd.ExecuteNonQuery() |> ignore

                    (* There should be only one row, so this closes the reader *)
                    minRdr.Read() |> ignore

let Reader _ =
    printfn "Reader process (using read-only transactions)"

    (* Create a connection for this process *)
    let conn = new RdmConnection("host=local;database=SQL14;openmode=s;tfstype=tfsr;readonlytransactions=1")
    conn.Open()

    (* Create a command to do the read *)
    let cmd = conn.CreateCommand()
    cmd.CommandText <- "SELECT MIN(id), MAX(id) FROM Table1"

    (* Loop until a shutdown is requested *)
    while true do
        (* Wait one second *)
        Thread.Sleep(1000)

        (* Query the minimum and maximum values *)
        let rdr = cmd.ExecuteReader()
        if rdr.Read() then
            printfn "Current minimum = %A and current minimum = %A" rdr.[0] rdr.[1]
            rdr.Read() |> ignore

let args = System.Environment.GetCommandLineArgs()
if args.Length > 1 then
    match args.[1] with
    | "inserter" -> Inserter ()
    | "deleter" -> Deleter ()
    | "reader" -> Reader ()
else
    printfn "Starting tfserver (and waiting 5 seconds for it to initialize)"

    (* Start the tfserver process to process 'remote' connections *)
    let tfserver = new Process()
    tfserver.StartInfo.FileName <- "tfserver.exe"
    tfserver.Start() |> ignore

    (* Sleep to give the tfserver time to initialize and start listening *)
    Thread.Sleep(5000)

    (* Create the connection with a simple connection string *)
    let conn = new RdmConnection("host=local;openmode=s;tfstype=tfsr")
    conn.Open()

    CreateDatabases conn

    (* Create three processes, one to insert, one to delete, and one
     * to show the current contents *)
    let inserter = new Process()
    inserter.StartInfo.FileName <- "SQL14_FS_ADO.exe"
    inserter.StartInfo.Arguments <- "inserter"

    let deleter = new Process()
    deleter.StartInfo.FileName <- "SQL14_FS_ADO.exe"
    deleter.StartInfo.Arguments <- "deleter"

    let reader = new Process()
    reader.StartInfo.FileName <- "SQL14_FS_ADO.exe"
    reader.StartInfo.Arguments <- "reader"

    printfn "Starting processes"
    printfn "   Press <Enter> when you are ready to terminate them"

    (* Start each of the processes *)
    inserter.Start() |> ignore
    deleter.Start() |> ignore
    reader.Start() |> ignore

    (* Wait for the user to hit <Enter> *)
    System.Console.ReadLine() |> ignore

    (* Shutdown the processes *)
    inserter.CloseMainWindow() |> ignore
    deleter.CloseMainWindow() |> ignore
    reader.CloseMainWindow() |> ignore

    (* Wait for each process to finish *)
    inserter.WaitForExit()
    deleter.WaitForExit()
    reader.WaitForExit()

    printfn "All processes have terminated"

    conn.Close()

    (* Stop the tfserver *)
    tfserver.CloseMainWindow() |> ignore
    tfserver.WaitForExit()
    tfserver.Close()

    printfn "Hit <Enter> to continue"
    System.Console.ReadLine() |> ignore

